@@ -20,7 +20,7 @@ import sys
 import tempfile
 from typing import Dict, List, Optional, Sequence, Union
 import warnings
-from .base import Converter, CodeChunk
+from .base import CodeChunk, Converter, Include
 from .. import err
 from .. import util
 
@@ -86,9 +86,20 @@ def _get_keyval_processors():
     '''
     Create dict mapping Pandoc key-value attributes to processing functions.
     '''
+    # Options like `include` have sub-options, which need to be translated
+    # into a dict.  In a markup language with more expressive option syntax,
+    # this would be supported as `include={key=value, ...}`.  For Pandoc
+    # Markdown, the alternatives are keys like `include.file` or
+    # `include_file`.  In either case, the base option name `include`
+    # functions as a namespace and needs to be protected from being
+    # overwritten by an invalid option.  Track those namespaces with a set.
+    namespace_keywords = set()
+
     def keyval_generic(code_chunk, options, key, value):
         if key in options:
             code_chunk.source_errors.append('Duplicate "{0}" attribute for code chunk'.format(key))
+        elif key in namespace_keywords:
+            code_chunk.source_errors.append('Invalid key "{0}" for code chunk'.format(key))
         else:
             options[key] = value
 
@@ -100,13 +111,24 @@ def _get_keyval_processors():
         else:
             options[key] = value == 'true'
 
+    def keyval_int(code_chunk, options, key, value):
+        if key in options:
+            code_chunk.source_errors.append('Duplicate "{0}" attribute for code chunk'.format(key))
+        else:
+            try:
+                value = int(value)
+            except (ValueError, TypeError):
+                code_chunk.source_errors.append('Attribute "{0}" must be integer for code chunk'.format(key))
+            else:
+                options[key] = value
+
     def keyval_first_number(code_chunk, options, key, value):
         if 'first_number' in options:
             code_chunk.source_warnings.append('Duplicate first line number attribute for code chunk')
         else:
             try:
                 value = int(value)
-            except ValueError:
+            except (ValueError, TypeError):
                 pass
             options['first_number'] = value
 
@@ -116,7 +138,7 @@ def _get_keyval_processors():
         elif value not in ('true', 'false'):
             code_chunk.source_warnings.append('Attribute "{0}" must be true or false for code chunk'.format(key))
         else:
-            options['line_anchors'] = value
+            options['line_anchors'] = value == 'true'
 
     def keyval_line_numbers(code_chunk, options, key, value):
         if 'line_numbers' in options:
@@ -124,20 +146,50 @@ def _get_keyval_processors():
         elif value not in ('true', 'false'):
             code_chunk.source_warnings.append('Attribute "{0}" must be true or false for code chunk'.format(key))
         else:
-            options['line_numbers'] = value
+            options['line_numbers'] = value == 'true'
 
+    def keyval_namespace(code_chunk, options, key, value):
+        if '.' in key:
+            namespace, sub_key = key.split('.', 1)
+        else:
+            namespace, sub_key = key.split('_', 1)
+        if namespace in options:
+            sub_options = options[namespace]
+        else:
+            sub_options = {}
+            options[namespace] = sub_options
+        if sub_key in sub_options:
+            code_chunk.source_errors.append('Duplicate "{0}" attribute for code chunk'.format(key))
+        else:
+            sub_options[sub_key] = value
+
+    expand_tabs = {k: keyval_bool for k in ['{0}_expand_tabs'.format(dsp) if dsp else 'expand_tabs'
+                                            for dsp in ('', 'markup', 'copied_markup', 'code', 'stdout', 'stderr')]}
     first_number = {k: keyval_first_number for k in ('first_number', 'startFrom', 'start-from', 'start_from')}
+    include = {'include_{0}'.format(k): keyval_namespace for k in Include.keywords}
     line_anchors = {k: keyval_line_anchors for k in ('lineAnchors', 'line-anchors', 'line_anchors')}
     line_numbers = {k: keyval_line_numbers for k in ('line_numbers', 'numberLines', 'number-lines', 'number_lines')}
+    rewrap_lines = {k: keyval_bool for k in ['{0}_rewrap_lines'.format(dsp) if dsp else 'rewrap_lines'
+                                             for dsp in ('', 'markup', 'copied_markup', 'code', 'stdout', 'stderr')]}
+    rewrap_width = {k: keyval_int for k in ['{0}_rewrap_width'.format(dsp) if dsp else 'rewrap_width'
+                                            for dsp in ('', 'markup', 'copied_markup', 'code', 'stdout', 'stderr')]}
+    tab_size = {k: keyval_int for k in ['{0}_tab_size'.format(dsp) if dsp else 'tab_size'
+                                         for dsp in ('', 'markup', 'copied_markup', 'code', 'stdout', 'stderr')]}
+    namespace_keywords.add('include')
     return collections.defaultdict(lambda: keyval_generic,
                                    {'complete': keyval_bool,
                                     'copy': keyval_generic,
                                     'example': keyval_bool,
+                                    **expand_tabs,
                                     **first_number,
-                                    'name': keyval_generic,
+                                    **include,
                                     **line_anchors,
                                     **line_numbers,
-                                    'outside_main': keyval_bool})
+                                    'name': keyval_generic,
+                                    'outside_main': keyval_bool,
+                                    **rewrap_lines,
+                                    **rewrap_width,
+                                    **tab_size})
 
 
 
@@ -167,8 +219,6 @@ class PandocCodeChunk(CodeChunk):
         options = {}
 
         # Preprocess options
-        if node_id:
-            options['name'] = node_id
         inline = node['t'] == 'Code'
         for n, c in enumerate(node_classes):
             self._class_processors[c](self, options, n, c)
@@ -191,14 +241,14 @@ class PandocCodeChunk(CodeChunk):
                     'codebraid_pseudonode' in inline_parent_node or len(parent_node_list) > 1)):
                 self.source_errors.insert(0, 'Option "example" is only allowed for inline code that is in a paragraph by itself')
                 self.options['example'] = False
-        pandoc_id = self.options.get('name', '')
+        pandoc_id = node_id
         pandoc_classes = []
         pandoc_kvpairs = []
         if 'lang' in options:
             pandoc_classes.append(options['lang'])
         if line_anchors:
             pandoc_classes.append('lineAnchors')
-        if self.options.get('line_numbers', False):
+        if self.options.get('code_line_numbers', False):
             pandoc_classes.append('numberLines')
         # Can't handle `startFrom` yet here, because if it is `next`, then
         # the value depends on which other code chunks end up in the session.
@@ -208,8 +258,8 @@ class PandocCodeChunk(CodeChunk):
         self.pandoc_classes = pandoc_classes
         self.pandoc_kvpairs = pandoc_kvpairs
         self._output_nodes = None
-        self._as_markdown = None
-        self._as_example_markdown = None
+        self._as_markup_lines = None
+        self._as_example_markup_lines = None
 
 
     _class_processors = _get_class_processors()
@@ -219,6 +269,12 @@ class PandocCodeChunk(CodeChunk):
     _unquoted_kv_value_re = re.compile(r'[A-Za-z$_+\-][A-Za-z0-9$_+\-:]*')
 
 
+    def finalize_after_copy(self):
+        if self.options['lang'] is None:
+            self.pandoc_classes.insert(0, self.copy_chunks[0].options['lang'])
+        self.options.finalize_after_copy()
+
+
     @property
     def output_nodes(self):
         '''
@@ -239,10 +295,10 @@ class PandocCodeChunk(CodeChunk):
                 nodes.append({'t': 'CodeBlock', 'c': [['', ['sourceError'], []], '{0}\n{1}'.format(message, '\n'.join(self.source_errors))]})
             self._output_nodes = nodes
             return nodes
-        if not self.inline and self.options['line_numbers']:
+        if not self.inline and self.options['code_line_numbers']:
             # Line numbers can't be precalculated since they are determined by
             # how a session is assembled across potentially multiple sources
-            first_number = self.options['first_number']
+            first_number = self.options['code_first_number']
             if first_number == 'next':
                 first_number = str(self.code_start_line_number)
             else:
@@ -251,97 +307,84 @@ class PandocCodeChunk(CodeChunk):
         t_code = 'Code' if self.inline else 'CodeBlock'
         t_raw = 'RawInline' if self.inline else 'RawBlock'
         for output, format in self.options['show'].items():
-            if output == 'markup':
-                nodes.append({'t': t_code, 'c': [['', ['markdown'], []], self.as_markdown]})
+            new_nodes = None
+            if output in ('markup', 'copied_markup'):
+                new_nodes = [{'t': t_code, 'c': [['', ['markdown'], []], self.layout_output(output, format)]}]
             elif output == 'code':
-                nodes.append({'t': t_code, 'c': [[self.pandoc_id, self.pandoc_classes, self.pandoc_kvpairs], self.code]})
-            elif output == 'expr':
+                new_nodes = [{'t': t_code, 'c': [[self.pandoc_id, self.pandoc_classes, self.pandoc_kvpairs], self.layout_output(output, format)]}]
+            elif output in ('expr', 'stdout', 'stderr'):
                 if format == 'verbatim':
-                    if self.expr_lines is not None:
-                        nodes.append({'t': t_code, 'c': [['', ['expr'], []], ' '.join(self.expr_lines)]})
+                    if getattr(self, output+'_lines') is not None:
+                        new_nodes = [{'t': t_code, 'c': [['', [output], []], self.layout_output(output, format)]}]
                 elif format == 'verbatim_or_empty':
-                    if self.expr_lines is not None:
-                        nodes.append({'t': t_code, 'c': [['', ['expr'], []], ' '.join(self.expr_lines)]})
-                    else:
-                        nodes.append({'t': t_code, 'c': [['', ['expr'], []], '\xa0']})
+                    new_nodes = [{'t': t_code, 'c': [['', [output], []], self.layout_output(output, format)]}]
                 elif format == 'raw':
-                    if self.expr_lines is not None:
-                        nodes.append({'t': t_raw, 'c': ['markdown', ' '.join(self.expr_lines)]})
-                else:
-                    raise ValueError
-            elif output == 'stdout':
-                if format == 'verbatim':
-                    if self.stdout_lines is not None:
-                        if self.inline:
-                            nodes.append({'t': t_code, 'c': [['', ['stdout'], []], ' '.join(self.stdout_lines)]})
-                        else:
-                            nodes.append({'t': t_code, 'c': [['', ['stdout'], []], '\n'.join(self.stdout_lines)]})
-                elif format == 'verbatim_or_empty':
-                    if self.stderr_lines is not None:
-                        if self.inline:
-                            nodes.append({'t': t_code, 'c': [['', ['stdout'], []], ' '.join(self.stdout_lines)]})
-                        else:
-                            nodes.append({'t': t_code, 'c': [['', ['stdout'], []], '\n'.join(self.stdout_lines)]})
-                    else:
-                        if self.inline:
-                            nodes.append({'t': t_code, 'c': [['', ['stdout'], []], '\xa0']})
-                        else:
-                            nodes.append({'t': t_code, 'c': [['', ['stdout'], []], '\xa0\n']})
-                elif format == 'raw':
-                    if self.stdout_lines is not None:
-                        if self.inline:
-                            nodes.append({'t': t_raw, 'c': ['markdown', ' '.join(self.stdout_lines)]})
-                        else:
-                            nodes.append({'t': t_raw, 'c': ['markdown', '\n'.join(self.stdout_lines)+'\n']})
-                else:
-                    raise ValueError
-            elif output == 'stderr':
-                if format == 'verbatim':
-                    if self.stderr_lines is not None:
-                        if self.inline:
-                            nodes.append({'t': t_code, 'c': [['', ['stderr'], []], ' '.join(self.stderr_lines)]})
-                        else:
-                            nodes.append({'t': t_code, 'c': [['', ['stderr'], []], '\n'.join(self.stderr_lines)]})
-                elif format == 'verbatim_or_empty':
-                    if self.stderr_lines is not None:
-                        if self.inline:
-                            nodes.append({'t': t_code, 'c': [['', ['stderr'], []], ' '.join(self.stderr_lines)]})
-                        else:
-                            nodes.append({'t': t_code, 'c': [['', ['stderr'], []], '\n'.join(self.stderr_lines)]})
-                    else:
-                        if self.inline:
-                            nodes.append({'t': t_code, 'c': [['', ['stderr'], []], '\xa0']})
-                        else:
-                            nodes.append({'t': t_code, 'c': [['', ['stderr'], []], '\xa0\n']})
-                elif format == 'raw':
-                    if self.stderr_lines is not None:
-                        if self.inline:
-                            nodes.append({'t': t_raw, 'c': ['markdown', ' '.join(self.stderr_lines)]})
-                        else:
-                            nodes.append({'t': t_raw, 'c': ['markdown', '\n'.join(self.stderr_lines)+'\n']})
+                    if getattr(self, output+'_lines') is not None:
+                        new_nodes = [{'t': t_raw, 'c': ['markdown', self.layout_output(output, format)]}]
                 else:
                     raise ValueError
+            elif output == 'rich_output':
+                if self.rich_output is None:
+                    continue
+                new_nodes = []
+                for ro in self.rich_output:
+                    ro_data = ro['data']
+                    ro_files = ro['files']
+                    for fmt in format:
+                        fmt_mime_type = self.options.mime_map[fmt]
+                        if fmt_mime_type not in ro_data:
+                            continue
+                        if fmt_mime_type in ro_files:
+                            image_node = {'t': 'Image', 'c': [['', [], []], [], [ro_files[fmt_mime_type], '']]}
+                            if self.inline:
+                                new_nodes.append(image_node)
+                            else:
+                                para_node = {'t': 'Para', 'c': [image_node]}
+                                new_nodes.append(para_node)
+                            break
+                        data = ro_data[fmt_mime_type]
+                        if fmt == 'latex':
+                            raw_node = {'t': 'RawInline', 'c': ['tex', data]}
+                            if self.inline:
+                                new_nodes.append(raw_node)
+                            else:
+                                para_node = {'t': 'Para', 'c': [raw_node]}
+                                new_nodes.append(para_node)
+                            break
+                        if fmt in ('html', 'markdown'):
+                            raw_node = {'t': t_raw, 'c': [fmt, data]}
+                            new_nodes.append(raw_node)
+                            break
+                        if fmt == 'plain':
+                            lines = util.splitlines_lf(data) or None
+                            if lines is not None:
+                                code_node = {'t': t_code, 'c': [['', [], []], self.layout_output(output, 'verbatim', lines)]}
+                                new_nodes.append(code_node)
+                            break
+                        raise ValueError
             else:
                 raise ValueError
+            if new_nodes is not None:
+                nodes.extend(new_nodes)
         self._output_nodes = nodes
         return nodes
 
 
     @property
-    def as_markdown(self):
-        if self._as_markdown is not None:
-            return self._as_markdown
-        self._as_markdown = self._generate_markdown()
-        return self._as_markdown
+    def as_markup_lines(self):
+        if self._as_markup_lines is not None:
+            return self._as_markup_lines
+        self._as_markup_lines = self._generate_markdown_lines()
+        return self._as_markup_lines
 
     @property
-    def as_example_markdown(self):
-        if self._as_example_markdown is not None:
-            return self._as_example_markdown
-        self._as_example_markdown = self._generate_markdown(example=True)
-        return self._as_example_markdown
+    def as_example_markup_lines(self):
+        if self._as_example_markup_lines is not None:
+            return self._as_example_markup_lines
+        self._as_example_markup_lines = self._generate_markdown_lines(example=True)
+        return self._as_example_markup_lines
 
-    def _generate_markdown(self, example=False):
+    def _generate_markdown_lines(self, example=False):
         '''
         Generate an approximation of the Markdown source that created the
         node.  This is used with `show=markup` and in creating examples that
@@ -352,15 +395,20 @@ class PandocCodeChunk(CodeChunk):
             attr_list.append('#{0}'.format(self.node_id))
         for c in self.node_classes:
             attr_list.append('.{0}'.format(c))
+        hide_keys = set(self.options.get('hide_markup_keys', []))
+        if example:
+            hide_keys.add('example')
         for k, v in self.node_kvpairs:
-            if not example or k != 'example':
+            if k not in hide_keys:
                 # Valid keys don't need quoting, some values may
                 if not self._unquoted_kv_value_re.match(v):
                     v = '"{0}"'.format(v.replace('\\', '\\\\').replace('"', '\\"'))
                 attr_list.append('{0}={1}'.format(k, v))
-        if self.placeholder_code is not None:
-            code = self.placeholder_code
+        if self.placeholder_code_lines is not None:
+            code_lines = self.placeholder_code_lines
+            code = code_lines[0]
         else:
+            code_lines = self.code_lines
             code = self.code
         if self.inline:
             code_strip = code.strip(' ')
@@ -371,15 +419,15 @@ class PandocCodeChunk(CodeChunk):
             delim = '`'
             while delim in code:
                 delim += '`'
-            md = '{delim}{code}{delim}{{{attr}}}'.format(delim=delim, code=code, attr=' '.join(attr_list))
-        elif self.placeholder_code is None or self.placeholder_code:
+            md_lines = ['{delim}{code}{delim}{{{attr}}}'.format(delim=delim, code=code, attr=' '.join(attr_list))]
+        elif self.placeholder_code_lines is None or code:
             delim = '```'
             while delim in code:
                 delim += '```'
-            md = '{delim}{{{attr}}}\n{code}\n{delim}'.format(delim=delim, attr=' '.join(attr_list), code=code)
+            md_lines = ['{delim}{{{attr}}}'.format(delim=delim, attr=' '.join(attr_list)), *code_lines, delim]
         else:
-            md = '```{{{attr}}}\n```'.format(attr=' '.join(attr_list))
-        return md
+            md_lines = ['```{{{attr}}}'.format(attr=' '.join(attr_list)), '```']
+        return md_lines
 
 
     def update_parent_node(self):
@@ -390,7 +438,7 @@ class PandocCodeChunk(CodeChunk):
             index = self.parent_node_list_index
             self.parent_node_list[index:index+1] = self.output_nodes
             return
-        markup_node = {'t': 'CodeBlock', 'c': [['', [], []], self.as_example_markdown]}
+        markup_node = {'t': 'CodeBlock', 'c': [['', [], []], self.layout_output('example_markup', 'verbatim')]}
         example_div_contents = [{'t': 'Div', 'c': [['', ['exampleMarkup'], []], [markup_node]]}]
         output_nodes = self.output_nodes
         if output_nodes:
@@ -609,7 +657,6 @@ class PandocConverter(Converter):
             to_format_pandoc_extensions = ''
         if input and input_paths:
             raise TypeError
-
         cmd_list = [str(self.pandoc_path),
                     '--from', from_format + from_format_pandoc_extensions]
         if newline_lf:
